
from excelexporter.generators import gds1
from excelexporter.generators import gds2
from excelexporter.generators import json
from excelexporter.generators import json2
from excelexporter.generators import resource

builtins = {
    "GDS1.0": gds1,
    "GDS2.0": gds2,
    "JSON1.0": json,
    "JSON2.0": json2,
    "RESOURCE": resource,
}



import glob
import json
import os
import logging
import textwrap
from excelexporter.config import Configuration
from excelexporter.sheetdata import SheetData


# 导出格式
extension = "json"

logger = logging.getLogger(__name__)


EE_DATATABLE_CLASS = """
class_name EEDataTable
extends Resource

@export
var data = {}
"""


def generator(sheetdata: SheetData, config: Configuration):
    # 表格数据脚本模板

    table = {}

    for id, row in sheetdata.items():
        row_data = {}

        for field, var in row.items():
            field_name: str = field
            row_data[field_name] = var.value

        table[id] = row_data

    code = json.dumps(table, ensure_ascii=False, indent=2)

    return code


def completed_hook(config: Configuration):

    output = config.output
    settings_file_path = os.path.join(output, "settings.gd")
    project_root = config.project_root

    lines = []

    loader = textwrap.dedent("""
    static func loader(path:String):
        var file = FileAccess.open(path,FileAccess.READ)
        var txt = file.get_as_text()
        var data = JSON.parse_string(txt)
        file.close()
        return data
    """)

    for path in glob.glob(f"{output}/**/*.json", recursive=True):
        if path == settings_file_path:
            continue  # 跳过 settings.gd
        basename = os.path.basename(path)
        setting_name = os.path.splitext(basename)[0]
        relpath = os.path.relpath(path, project_root).replace("\\", "/")
        lines.append(f"var {setting_name} = loader('res://{relpath}')")

        # 去掉缩进
    code = textwrap.dedent(
        """
    extends Node
    # 这个脚本你需要挂到游戏的Autoload才能全局读表

    {loader}
    {refs_code}
    """
    )
    refs_code = "\n".join(lines)

    code = code.format(loader=loader, refs_code=refs_code)

    with open(settings_file_path, "w", encoding="utf-8", newline="\n") as f:
        f.write(code)
        logger.info("创建setting.gd")



from dataclasses import dataclass
import logging

from excelexporter.sheetdata import SheetData, TypeDefine
from excelexporter.config import Configuration
from typing import Any, Callable


Generator = Callable[[SheetData, Configuration], str]
CompletedHook = Callable[[Configuration], None]

ConvertFunc = Callable[[Any, str, int, dict], str]

logger = logging.getLogger()


@dataclass
class Variant:

    id: int
    type_define: TypeDefine
    field_name: str
    value: Any

    def local_strs(self):
        return set()


class String(Variant):
    @staticmethod
    def make(id: Any, td: TypeDefine, fn: str, v: Any):
        value = str(v) if v else ""
        return String(id, td, fn, value)

    def local_strs(self):
        localizeds = set()
        if self.type_define.is_localization:
            localizeds.add(self.value)
        return localizeds


class Int(Variant):
    @staticmethod
    def make(id: Any, td: TypeDefine, fn: str, v: Any):
        value = int(float(v or 0))
        return Int(id, td, fn, value)


class Float(Variant):
    @staticmethod
    def make(id: Any, td: TypeDefine, fn: str, v: Any):
        value = float(v or 0)
        return Float(id, td, fn, value)


class Bool(Variant):
    @staticmethod
    def make(id: Any, td: TypeDefine, fn: str, v: Any):
        value = v != "FALSE"
        return Bool(id, td, fn, value)


class Array(Variant):
    @staticmethod
    def make(id: Any, td: TypeDefine, fn: str, v: Any):
        value = eval(f'[{v.replace("|",",")}]') if v else []
        return Array(id, td, fn, value)

    def local_strs(self):
        localizeds = set()
        if self.type_define.is_localization:
            for e in self.value:
                if isinstance(e, str):
                    localizeds.add(e)
        return localizeds


class ArrayStr(Variant):
    @staticmethod
    def make(id: Any, td: TypeDefine, fn: str, v: Any):
        value = ["%s" % e for e in v.split("|")]if v else []
        return ArrayStr(id, td, fn, value)

    def local_strs(self):
        localizeds = set()
        if self.type_define.is_localization:
            for e in self.value:
                if isinstance(e, str):
                    localizeds.add(e)
        return localizeds


class ArrayBool(Variant):

    @staticmethod
    def make(id: Any, td: TypeDefine, fn: str, v: Any):
        value = [e != "FALSE" for e in v.split("|")] if v else []
        return ArrayBool(id, td, fn, value)


class Dict(Variant):
    @staticmethod
    def make(id: Any, td: TypeDefine, fn: str, v: Any):
        value = eval(f'{{{v.replace("|",",")}}}') if v else {}
        return Dict(id, td, fn, value)

    def local_strs(self):
        localizeds = set()
        if self.type_define.is_localization:
            for k, e in self.value.items():
                if isinstance(e, str):
                    localizeds.add(e)
        return localizeds


class Type:
    ID = "id"
    STRING = "string"
    INT = "int"
    FLOAT = "float"
    BOOL = "bool"
    ARRAY = "array"
    ARRAY_STR = "array_str"
    ARRAY_BOOL = "array_bool"
    DICT = "dict"
    FUNCTION = "function"


class Converter:
    """
    字段转换器，将excel字段值转python数据
    """

    def __init__(self) -> None:
        self._map = {}
        self.functions: list[Variant] = []

        self.register(Type.ID, Int.make)
        self.register(Type.STRING, String.make)
        self.register(Type.INT, Int.make)
        self.register(Type.FLOAT, Float.make)
        self.register(Type.BOOL, Bool.make)
        self.register(Type.ARRAY, Array.make)
        self.register(Type.ARRAY_STR, ArrayStr.make)
        self.register(Type.ARRAY_BOOL, ArrayBool.make)
        self.register(Type.DICT, Dict.make)
        self.register(Type.FUNCTION, Variant)

    def default(self, id, td, fn, value): return value or 0

    def register(self, type: str, self_method: ConvertFunc):
        self._map[type] = self_method

    def __call__(
            self,
            id: int,
            type_define: TypeDefine,
            field_name: str,
            value: Any,
            *args: Any,
            **kwds: Any
    ) -> Any:
        cvt = self._map.get(type_define.type_name, self.default)
        result = cvt(id, type_define, field_name, value)
        return result



import logging
import toml
from dataclasses import dataclass, field, asdict

logger = logging.getLogger(__name__)


@dataclass
class Localization:
    babel_keywords: list = field(
        default_factory=lambda: ["tr", "Label/text"])
    pot_file: str = "lang/template.pot"


@dataclass
class Configuration:
    ignore_sheet_mark: str = field(default="~")
    ignore_field_mark: str = field(default="*")
    custom_generator: str = field(default="GDS2.0")
    input: str = field(default="data")
    output: str = field(default="dist")
    project_root: str = field(default="../")
    localization: Localization = field(default_factory=Localization)

    @classmethod
    def load(cls, filename: str = "export.toml") -> 'Configuration':
        with open(filename) as f:
            data = toml.load(f)
            config = Configuration(**data)
            return config

    def save(self, filename: str = "export.toml"):
        with open(filename, "w") as f:
            data = asdict(self)
            toml.dump(data, f)



import re
from dataclasses import dataclass, field
from typing import List


@dataclass
class TypeDefine:
    """
    # string(params)
    functioN(params)
    [#]<type_name>[(params)]

    字段转换器要如何使用这些属性字段全靠他们自己定义了
    """
    TYPE_DEFINE_PATTERN = r"(?P<local>#?)(?P<type_name>\w+)(?P<params>\(.*\))?"
    is_localization: bool  # 是否需要支持多语言，由具体的Parser去决定如何支持
    type_name: str
    params: str  # params是(a,b,c=null) 这样的字符串

    @classmethod
    def from_str(self, define_str: str) -> 'TypeDefine':
        type_define = define_str.strip()  # 先去头去尾空格
        m = re.match(self.TYPE_DEFINE_PATTERN, type_define)
        is_localization = m.group("local")
        type_name = m.group("type_name")
        params = m.group("params") or "(args)"
        return TypeDefine(
            is_localization,
            type_name,
            params
        )


@dataclass
class Define:
    type: List[TypeDefine] = field(default_factory=list)
    desc: List[str] = field(default_factory=list)
    name: List[str] = field(default_factory=list)


@dataclass
class SheetData:
    define: Define = field(default_factory=Define)
    table: list = field(default=list)



import logging
import sys

FORMAT = "%(name)s-%(levelname)s:%(message)s"

fh = logging.FileHandler("log.txt", "w")
fh.setLevel(logging.DEBUG)
fh.setFormatter(logging.Formatter(FORMAT))

ch = logging.StreamHandler(sys.stdout)
ch.setLevel(logging.DEBUG)
ch.setFormatter(logging.Formatter(FORMAT))

logging.basicConfig(level=logging.DEBUG, handlers=[fh, ch])


def log_uncaught_exceptions(ex_cls, ex, tb):
    logging.critical("未捕获异常", exc_info=(ex_cls, ex, tb))


sys.excepthook = log_uncaught_exceptions



import glob
import os
import pprint
import logging
import textwrap
import re
from excelexporter.config import Configuration
from excelexporter.sheetdata import SheetData


# 导出格式
extension = "tres"

logger = logging.getLogger(__name__)


EE_DATATABLE_CLASS = """
class_name EEDataTable
extends Resource

@export
var data = {}
"""


def generator(sheetdata: SheetData, config: Configuration):
    # 表格数据脚本模板
    abs_output = os.path.abspath(config.output)
    relpath = os.path.relpath(
        abs_output, config.project_root).replace("\\", "/")
    template = """
    [gd_resource type="Resource" script_class="EEDataTable" load_steps=2 format=3] 

    [ext_resource type="Script" path="res://{relpath}/ee_data_table.gd" id="1"]

    [resource]
    script = ExtResource("1")
    data = {data}
    """  # noqa
    template = textwrap.dedent(template)
    table = {}

    for id, row in sheetdata.items():
        row_data = {}

        for field, var in row.items():
            field_name: str = field
            row_data[field_name] = var.value

        table[id] = row_data

    code = template.format(
        data=pprint.pformat(
            table, indent=2, width=1000000000,
            compact=True, sort_dicts=True),
        relpath=relpath
    )

    code = textwrap.dedent(code)
    code = re.sub(r"\b(True|False)\b", lambda m: m.group(1).lower(), code)
    code = code.replace("'", '"')
    return code


def completed_hook(config: Configuration):

    output = config.output
    settings_file_path = os.path.join(output, "settings.gd")
    data_class_file_path = os.path.join(output, "ee_data_table.gd")
    project_root = config.project_root

    lines = []

    for path in glob.glob(f"{output}/**/*.{extension}", recursive=True):
        if path == settings_file_path:
            continue  # 跳过 settings.gd
        if path == data_class_file_path:
            continue  # 跳过数据表类
        basename = os.path.basename(path)
        setting_name = os.path.splitext(basename)[0]
        relpath = os.path.relpath(path, project_root).replace("\\", "/")
        lines.append(f"var {setting_name} = load('res://{relpath}')")

        # 去掉缩进
    code = textwrap.dedent(
        """
    extends Node
    # 这个脚本你需要挂到游戏的Autoload才能全局读表

    {refs_code}
    """
    )
    refs_code = "\n".join(lines)

    code = code.format(refs_code=refs_code)

    with open(settings_file_path, "w", encoding="utf-8", newline="\n") as f:
        f.write(code)
        logger.info("创建setting.gd")

    with open(data_class_file_path, "w", encoding="utf-8", newline="\n") as f:
        f.write(EE_DATATABLE_CLASS)
        logger.info("创建EEDataTable类")



from babel.messages.jslexer import tokenize, unquote_string


# This helper is copied to all necessary files because
# once again python imports are being a pain
def reopen_normal_read(file_obj, encoding):
    """Re-open a file obj in plain read mode"""
    return open(file_obj.name, "r", encoding=encoding)


class ActiveCall:
    def __init__(self, name, parenthesis_level):
        self.name = name
        self.current_value = None
        self.value_start_line = 0
        self.stored_parenthesis_level = parenthesis_level

    def valid(self):
        return self.name and self.current_value


class CSharpExtractor(object):
    def __init__(self, data):
        self.data = data

        self.current_name = None
        self.active_calls = []

        self.parenthesis_level = 0

        self.results = []

    def start_call(self):
        self.active_calls.append(ActiveCall(
            self.current_name, self.parenthesis_level))
        self.parenthesis_level = 0
        self.current_name = None

    def end_call(self):
        call = self.active_calls.pop()

        self.parenthesis_level = call.stored_parenthesis_level

        if call.valid():
            self.add_result(call)

    def add_result(self, call):
        result = dict(
            line_number=call.value_start_line,
            content=call.current_value,
            function_name=call.name
        )
        self.results.append(result)

    def get_lines_data(self):
        """
        Returns string:line_numbers list
        Since all strings are unique it is OK to get line numbers this way.
        :rtype: list
        """
        trigger_call_prime = False

        for token in tokenize(self.data, jsx=False):
            call_primed = trigger_call_prime
            trigger_call_prime = False

            if token.type == 'operator':
                if token.value == '(':
                    if call_primed:
                        self.start_call()
                    else:
                        self.parenthesis_level += 1
                elif token.value == ')':
                    if self.parenthesis_level == 0:
                        self.end_call()
                    else:
                        self.parenthesis_level -= 1
            elif token.type == 'name':
                trigger_call_prime = True
                self.current_name = token.value
            elif token.type == 'string' and len(self.active_calls) > 0:
                string_value = unquote_string(token.value)

                call = self.active_calls[-1]

                if call.current_value is None:
                    call.current_value = string_value
                    call.value_start_line = token.lineno
                else:
                    call.current_value += string_value

        return self.results


def extract_godot_csharp(file_obj, keywords, comment_tags, options):
    """
    Custom C# extract to fix line numbers for Windows
    """
    with reopen_normal_read(file_obj, options.get('encoding', 'utf-8')) as f:
        data = f.read()
    extractor = CSharpExtractor(data)

    for item in extractor.get_lines_data():
        function = item['function_name']

        if function not in keywords:
            continue

        messages = [item['content']]
        yield item['line_number'], function, tuple(messages), []



import glob
import os
import sys
import xlwings as xw
import logging
from excelexporter.config import Configuration
from excelexporter.generator import Converter, Generator, CompletedHook, Variant  # noqa
from excelexporter.sheetdata import SheetData, TypeDefine
from typing import Dict, Optional

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points

# 导表工具引擎

logger = logging.getLogger(__name__)


class IllegalFile(Exception):

    def __init__(self, filename: str, *args: object) -> None:
        super().__init__(f"{filename} 不是配置表目录下的配置", *args)


class IllegalGenerator(Exception):
    def __init__(self, name: str, *args: object) -> None:
        super().__init__(
            name,
            *args
        )


def discover_generator():
    generators = entry_points(
        group="excelexporter.generator")
    return generators


class Engine(xw.App):

    def __init__(self, config: Configuration) -> None:
        super().__init__(visible=False)
        self.config = config
        self.generator: Optional[Generator] = None
        self.completed_hook: Optional[CompletedHook] = None
        self.extension: str = ""

        self.localized_strs = set()
        self.cvt = Converter()

        self.init_generator()

    def init_generator(self):

        generator = None
        completed_hook = None
        extension = None

        # 如果有自定义导出器就优先用自定义导出器
        if self.config.custom_generator.endswith(".py"):
            with open(self.config.custom_generator) as f:
                code = f.read()
                exec(code)
                self.generator = generator
                self.completed_hook = completed_hook
                self.extension = extension

                if (generator and completed_hook and extension) is False:
                    raise IllegalGenerator(
                        self.config.custom_generator,
                        "自定义导出器不完整，generator、completed_hook、extension存在没定义。")

                logger.info(f"使用 {self.config.custom_generator} 自定义导出器")
        else:
            # 找出插件
            generators = discover_generator()

            if self.config.custom_generator in generators.names:
                module = generators[self.config.custom_generator].load()
                logger.info(
                    f"使用插件导出器 {self.config.custom_generator} :{module.__name__}")  # noqa
            else:
                raise IllegalGenerator(
                    self.config.custom_generator, generators.names)

            self.generator = getattr(module, "generator")
            self.completed_hook = getattr(module, "completed_hook")
            self.extension = getattr(module, "extension")

    def _gen(self, excel_file: str):
        wb_abs_path: str = os.path.abspath(excel_file)
        abs_input_path: str = os.path.abspath(self.config.input)
        abs_output_path: str = os.path.abspath(self.config.output)
        wb_abs_path_without_ext: str = os.path.splitext(wb_abs_path)[0]

        if not wb_abs_path.startswith(abs_input_path):
            raise IllegalFile(wb_abs_path, abs_input_path)

        sheet_datas = self._excel2dict(wb_abs_path)

        for sheet_name, sheetdata in sheet_datas.items():
            try:
                if "-" in sheet_name:
                    org_name, rename = sheet_name.split("-")
                else:
                    org_name, rename = sheet_name, None

                relative_path = os.path.join(
                    wb_abs_path_without_ext.replace(abs_input_path, ""),
                    rename or org_name
                )
                output = abs_output_path + relative_path

                output_dirname = os.path.dirname(output)
                # 保持输出的文件目录层级结构与输入一致
                if not os.path.exists(output_dirname):
                    os.makedirs(output_dirname)

                code = self.generator(sheetdata, self.config)

                # code = "# 本文件由代码生成，不要手动修改\n"+code
                output = f"{output}.{self.extension}"
                with open(output, "w", encoding="utf-8", newline="\n") as f:
                    f.write(code)
                    logger.info(f"导出：{wb_abs_path}:{sheet_name} => {output}")
            except Exception:
                logger.error(f"{sheet_name} 导出失败", exc_info=True)

    def _excel2dict(self, wb_file: str) -> Dict[str, SheetData]:
        """
        workbook解析加工成字典
        """
        with self.books.open(wb_file) as book:
            ignore_sheet_mark = self.config.ignore_sheet_mark
            # 过滤掉打了忽略标志的sheet
            sheets = filter(
                lambda sheet: not sheet.name.startswith(ignore_sheet_mark),
                book.sheets
            )

            wb_data = {}

            # 先讲sheet转sheet_data
            for sheet in sheets:
                sheet_data = SheetData()
                row_values = sheet.range("A1").expand().raw_value

                sheet_data.define.type = list(row_values[0])
                sheet_data.define.desc = list(row_values[1])
                sheet_data.define.name = list(row_values[2])

                sheet_data.table = list([list(row) for row in row_values[3:]])
                # 找出所有被打了忽略标记的字段
                for col, field in enumerate(sheet_data.define.name):
                    # 跳过没命令的字段

                    if field is None or field.startswith(self.config.ignore_field_mark):  # noqa
                        del sheet_data.define.type[col]
                        del sheet_data.define.desc[col]
                        del sheet_data.define.name[col]
                        for row in sheet_data.table:
                            del row[col]

                wb_data[sheet.name] = sheet_data

            cvt = Converter()
            for sheet_name, sheet_data in wb_data.items():
                field_names = sheet_data.define.name
                field_types = sheet_data.define.type
                table = {}

                for row in sheet_data.table:
                    id_type = TypeDefine.from_str(field_types[0])
                    id_name = field_names[0]
                    id_value = row[0]
                    id = cvt(id_value, id_type, id_name, id_value)

                    row_data = {}

                    for index, value in enumerate(row):
                        field_name: str = field_names[index]
                        field_type = TypeDefine.from_str(field_types[index])
                        variant: Variant = cvt(
                            id.value, field_type, field_name, value
                        )
                        row_data[field_name] = variant
                        self.localized_strs = self.localized_strs.union(
                            variant.local_strs())

                    table[id.value] = row_data
                wb_data[sheet_name] = table

            return wb_data

    def save_lang_file(self):

        with open("language.gd", "w", encoding="utf-8", newline="\n") as f:
            f.write("func localization():\n")
            lines = []
            for txt in self.localized_strs:
                lines.append(
                    f"  tr('{txt}')\n"
                )
            f.writelines(lines)

    def gen_one(self, filename: str):
        self._gen(filename)
        if self.completed_hook:
            self.completed_hook(self.config)

    def gen_all(self):
        abs_input = os.path.abspath(self.config.input)
        exts = [".xlsx", ".xls"]
        for ext in exts:
            full_paths = glob.glob(f"{abs_input}/**/*{ext}", recursive=True)
            for full_path in full_paths:
                filename = os.path.basename(full_path)
                if filename.startswith("~$"):
                    logger.warning(f"{filename} 不是配置表，跳过！")
                    continue
                self._gen(full_path)

        if self.completed_hook:
            self.completed_hook(self.config)

    def extract_pot(self):
        abs_input = os.path.abspath(self.config.input)
        exts = [".xlsx", ".xls"]
        for ext in exts:
            full_paths = glob.glob(f"{abs_input}/**/*{ext}", recursive=True)
            for full_path in full_paths:
                filename = os.path.basename(full_path)
                if filename.startswith("~$"):
                    logger.warning(f"{filename} 不是配置表，跳过！")
                    continue
                self._excel2dict(full_path)  # 直接读所有表，不做转换抽取翻译字符
                logger.info(f"导出语言表: {full_path}")
        self.save_lang_file()



import glob
import os
import logging
import jinja2
import pkg_resources
from excelexporter.config import Configuration
from excelexporter.sheetdata import SheetData
from excelexporter.generator import Variant
from excelexporter.generator import Type

# jinja2 docs: http://doc.yonyoucloud.com/doc/jinja2-docs-cn/templates.html#id2

logger = logging.getLogger(__name__)

# 导出格式
extension = "gd"

loader = jinja2.FileSystemLoader(
    pkg_resources.resource_filename(__package__, ""))
env = jinja2.Environment(autoescape=False, loader=loader)


def converter(var: Variant):
    type_define = var.type_define
    if type_define.type_name == Type.STRING:
        value = var.value.replace("\n", "\\n")
        return f"'{value}'"

    if type_define.type_name == Type.FUNCTION:
        func_name = f"{var.field_name}_{var.id}"
        return f"funcref(self,'{func_name}')"

    return var.value


env.filters["cvt"] = converter


def generator(sheetdata: SheetData, config: Configuration):
    # 表格数据脚本模板
    template = env.get_template("data_template.gd")
    code = template.render(sheetdata=sheetdata)
    return code


def completed_hook(config: Configuration):
    output = config.output
    settings_file_path = os.path.join(output, "settings.gd")
    project_root = config.project_root

    lines = []

    for path in glob.glob(f"{output}/**/*.{extension}", recursive=True):
        if path == settings_file_path:
            continue  # 跳过 settings.gd
        basename = os.path.basename(path)
        setting_name = os.path.splitext(basename)[0]
        relpath = os.path.relpath(path, project_root).replace("\\", "/")
        lines.append(f"var {setting_name} = load('res://{relpath}').new()")

    template = env.get_template("setting_template.gd")
    code = template.render(lines=lines)

    with open(settings_file_path, "w", encoding="utf-8", newline="\n") as f:
        f.write(code)
        logger.info("创建setting.gd")



from collections import deque
from babel.messages.jslexer import tokenize, unquote_string


JSON_GETTEXT_KEYWORD = 'type'
JSON_GETTEXT_VALUE = 'gettext_string'
JSON_GETTEXT_KEY_CONTENT = 'content'
JSON_GETTEXT_KEY_ALT_CONTENT = 'alt_content'
JSON_GETTEXT_KEY_FUNCNAME = 'funcname'


class JsonExtractor(object):
    def __init__(self, data):
        self.state = 'start'
        self.data = data

        self.token_to_add = None
        self.is_value = False
        self.gettext_mode = False
        self.current_key = None
        self.results = []
        self.token_params = {}

    def add_result(self, token):
        value = unquote_string(token.value)
        if value not in self.results:
            self.results[value] = deque()
        self.results[value].append(token.lineno)

    def start_object(self):
        self.gettext_mode = False
        self.state = 'key'

    def with_separator(self, token):
        self.state = 'value'

    def end_pair(self, add_gettext_object=False):
        if self.token_to_add:
            if not self.gettext_mode or (self.gettext_mode and add_gettext_object):
                self.add_result(self.token_to_add)

        self.current_key = None
        self.state = 'key'

    def end_object(self):
        self.end_pair(add_gettext_object=True)
        self.gettext_mode = False
        self.state = 'end'

    def add_result(self, token):
        value = unquote_string(token.value)
        result = dict(
            line_number=token.lineno,
            content=value
        )
        for key, value in self.token_params.items():
            if key == 'alt_token':
                result['alt_content'] = unquote_string(value.value)
                result['alt_line_number'] = value.lineno
            else:
                result[key] = unquote_string(value)

        self.results.append(result)
        self.token_to_add = None
        self.token_params = {}

    def get_lines_data(self):
        """
        Returns string:line_numbers list
        Since all strings are unique it is OK to get line numbers this way.
        Since same string can occur several times inside single .json file the values should be popped(FIFO) from the list
        :rtype: list
        """

        encoding = 'utf-8'

        for token in tokenize(self.data.decode(encoding)):
            if token.type == 'operator':
                if token.value == '{':
                    self.start_object()
                elif token.value == ':':
                    self.with_separator(token)
                elif token.value == '}':
                    self.end_object()
                elif token.value == ',':
                    self.end_pair()

            elif token.type == 'string':
                if self.state == 'key':
                    self.current_key = unquote_string(token.value)
                    if self.current_key == JSON_GETTEXT_KEYWORD:
                        self.gettext_mode = True

                # ==value not actually used, but if only key was met (like in list) it still will be used. The important part, that key wont be parsed as value, not reversal
                if self.gettext_mode:
                    if self.current_key == JSON_GETTEXT_KEY_CONTENT:
                        self.token_to_add = token
                    elif self.current_key == JSON_GETTEXT_KEY_ALT_CONTENT:
                        self.token_params['alt_token'] = token
                    elif self.current_key == JSON_GETTEXT_KEY_FUNCNAME:
                        self.token_params['funcname'] = token.value
                else:
                    self.token_to_add = token

        return self.results


def extract_godot_json(fileobj, keywords, comment_tags, options):
    """
    Supports: gettext, ngettext. See package README or github ( https://github.com/tigrawap/pybabel-json ) for more usage info.
    """
    data = fileobj.read()
    json_extractor = JsonExtractor(data)
    strings_data = json_extractor.get_lines_data()

    for item in strings_data:
        messages = [item['content']]
        if item.get('funcname') == 'ngettext':
            messages.append(item['alt_content'])
        yield item['line_number'], item.get('funcname', 'gettext'), tuple(messages), []



import shutil
import click
import logging
import os
import pkg_resources
from excelexporter.config import Configuration
from excelexporter.engine import Engine, discover_generator

logger = logging.getLogger(__name__)


@click.group
def main():
    """
    ===============================
    Excel表导出工具
    ===============================
    """
    pass


@main.command
@click.option(
    "--setting-dir", "-d",
    is_flag=True,
    default=True,
    prompt="创建并放在Setting目录？"
)
def init(setting_dir: bool):
    """
    生成默认配置表项目
    """
    setting_dir_name = "Setting"

    config = Configuration()
    input_dir = click.prompt(
        "输入存放excel表格目录名称", default=config.input, show_default=True)
    output_dir = click.prompt(
        "输入存放导出文件目录名称", default=config.output, show_default=True)

    template = pkg_resources.resource_filename(
        __package__,
        "template"
    )

    generator = click.prompt(
        "使用哪个内置导出器？",
        type=click.Choice(discover_generator().names),
        default="GDS2.0"
    )

    if setting_dir:
        if os.path.exists(setting_dir_name) and os.listdir(setting_dir_name):
            click.echo(f"{setting_dir_name} 已经存在并且非空!")
            return
        os.mkdir(setting_dir_name)
        os.chdir(setting_dir_name)

    if os.path.exists(input_dir) and os.listdir(input_dir):
        click.echo(f"{input_dir}已经存在并且非空!")
        return
    if os.path.exists(output_dir) and os.listdir(output_dir):
        click.echo(f"{output_dir}已经存在并且非空!")
        return

    config.input = input_dir
    config.output = output_dir

    os.mkdir(input_dir)
    os.mkdir(output_dir)
    os.mkdir("lang")
    config.custom_generator = generator
    config.save()

    shutil.copytree(template, os.curdir, dirs_exist_ok=True)
    click.echo("配置表项目生成完毕，后续你可以通过修改export.toml调整配置。")


@main.command(name="list")
def _list():
    """
    列出支持的导出器插件
    """
    generators = discover_generator()
    if generators.names:
        for gen in generators.names:
            print(gen)


@ main.command
def add_context_menu():
    """
    添加上下文菜单（通过注册表）
    """
    dir = pkg_resources.resource_filename(__package__, "template/reg")
    os.system(f"start {dir}")


def _find_config():
    if not os.path.exists("export.toml"):
        logger.error("当前目录下没有export.toml配置文件")
        logger.error("尝试往上层找")
        upper_path = os.path.join(os.curdir, os.pardir, "export.toml")
        if not os.path.exists(upper_path):
            logger.error("完全不存在export.toml,终止导表")
            raise FileNotFoundError("完全不存在export.toml,终止导表")
        else:
            os.chdir(os.pardir)


@ main.command
@ click.option("--cwd", default=".", help="工作目录，执行命令所在的目录")
def gen_all(cwd):
    """
    导出所有表
    """
    os.chdir(cwd)
    _find_config()

    config = Configuration.load()

    with Engine(config) as engine:
        engine.gen_all()


@ main.command()
@ click.argument("file", type=click.Path(True))
def gen_one(file: str):
    """
    打开并导出整张excel表
    """
    abs_filepath = os.path.abspath(file)
    _find_config()

    config = Configuration.load()
    with Engine(config) as engine:
        engine.gen_one(abs_filepath)


@ main.command
@ click.option("--cwd", default=".", help="工作目录，执行命令所在的目录")
def extract(cwd):
    """
    导出多语言表 gd,供 babel生成语言表用
    """
    os.chdir(cwd)
    _find_config()
    config = Configuration.load()
    with Engine(config) as engine:
        engine.extract_pot()

    babel_keywords = config.localization["babel_keywords"]
    pot_file = config.localization["pot_file"]

    keyword_args = "".join([f"-k {kw} " for kw in babel_keywords])
    os.system(
        f"pybabel extract -F babel.cfg {keyword_args} -o {pot_file} {config.project_root}"  # noqa
    )


if __name__ == "__main__":
    main()



import re

# thanks: https://github.com/remram44/pybabel-godot/edit/master/babel_godot.py

__version__ = '1.2'


_godot_node = re.compile(r'^\[node name="([^"]+)" (?:type="([^"]+)")?')
_godot_property_str = re.compile(
    r'^([A-Za-z0-9_]+)\s*=\s*([\["].+)\Z',
    re.DOTALL,
)


class StringReader(object):
    def __init__(self, lineno):
        self.result = []
        self.lineno = lineno

    def parse_line(self, string):
        escaped = False
        for i, c in enumerate(string):
            if escaped:
                if c == '\\':
                    self.result.append('\\')
                elif c == 'n':
                    self.result.append('\n')
                elif c == 't':
                    self.result.append('\t')
                else:
                    self.result.append(c)
                escaped = False
            else:
                if c == '\\':
                    escaped = True
                elif c == '"':
                    return string[i + 1:]
                else:
                    self.result.append(c)
        return None

    def get_result(self):
        return [(''.join(self.result), self.lineno)]


class ArrayReader(object):
    def __init__(self, lineno):
        self.result = []
        self.string = None
        self.lineno = lineno

    def parse_line(self, string):
        lineno = self.lineno
        self.lineno += 1

        if self.string is not None:
            remainder = self.string.parse_line(string)
            if remainder is None:
                return None
            self.result.extend(self.string.get_result())
            string = remainder
            self.string = None

        i = 0
        while i < len(string):
            c = string[i]
            if c in ' \t,':
                i = i + 1
            elif c == ']':
                return string[i + 1:]
            elif c == '"':
                self.string = StringReader(lineno)
                remainder = self.string.parse_line(string[i + 1:])
                if remainder is None:
                    return None
                else:
                    self.result.extend(self.string.get_result())
                    string = remainder
                    self.string = None
                    i = 0
            else:
                raise ValueError("Unexpected character %r" % (c,))

        raise ValueError("Unterminated array")

    def get_result(self):
        return self.result


def extract_godot_scene(fileobj, keywords: list, comment_tags, options):
    """Extract messages from Godot scene files (.tscn).

    :param fileobj: the seekable, file-like object the messages should be
                    extracted from
    :param keywords: a list of property names that should be localized, in the
                     format '<NodeType>/<name>' or '<name>' (example:
                     'Label/text')
    :param comment_tags: a list of translator tags to search for and include
                         in the results (ignored)
    :param options: a dictionary of additional options (optional)
    :rtype: iterator
    """
    encoding = options.get('encoding', 'utf-8')

    current_node_type = None

    properties_to_translate = {}
    for keyword in keywords:
        if '/' in keyword:
            properties_to_translate[tuple(keyword.split('/', 1))] = keyword
        else:
            properties_to_translate[(None, keyword)] = keyword

    def check_translate_property(property):
        keyword = properties_to_translate.get((current_node_type, property))
        if keyword is None:
            keyword = properties_to_translate.get((None, property))
        return keyword

    current_value = keyword = None

    for lineno, line in enumerate(fileobj, start=1):
        line = line.decode(encoding)

        if current_value:
            remainder = current_value.parse_line(line)
            if remainder is None:  # Still un-terminated
                pass
            elif remainder.strip():
                raise ValueError("Trailing data after string")
            else:
                for value, value_lineno in current_value.get_result():
                    yield (
                        value_lineno,
                        keyword,
                        [value],
                        [],
                    )
                current_value = None
            continue

        match = _godot_node.match(line)
        if match:
            # Store which kind of node we're in
            current_node_type = match.group(2)
            # Instanced packed scenes don't have the type field,
            # change current_node_type to empty string
            current_node_type = current_node_type \
                if current_node_type is not None else ""
        elif line.startswith('['):
            # We're no longer in a node
            current_node_type = None
        elif current_node_type is not None:
            # Currently in a node, check properties
            match = _godot_property_str.match(line)
            if match:
                property = match.group(1)
                value = match.group(2)
                keyword = check_translate_property(property)
                if keyword:
                    if value[0:1] == '[':
                        current_value = ArrayReader(lineno)
                    else:
                        current_value = StringReader(lineno)
                    remainder = current_value.parse_line(value[1:])
                    if remainder is None:
                        pass  # Un-terminated string
                    elif not remainder.strip():
                        for value, value_lineno in current_value.get_result():
                            yield (value_lineno, keyword, [value], [])
                        current_value = None
                    else:
                        raise ValueError("Trailing data after string")


def extract_godot_resource(fileobj, keywords, comment_tags, options):
    """Extract messages from Godot resource files (.res, .tres).

    :param fileobj: the seekable, file-like object the messages should be
                    extracted from
    :param keywords: a list of property names that should be localized, in the
                     format 'Resource/<name>' or '<name>' (example:
                     'Resource/text')
    :param comment_tags: a list of translator tags to search for and include
                         in the results (ignored)
    :param options: a dictionary of additional options (optional)
    :rtype: iterator
    """
    encoding = options.get('encoding', 'utf-8')

    properties_to_translate = {}
    for keyword in keywords:
        if keyword.startswith('Resource/'):
            properties_to_translate[keyword[9:]] = keyword

    def check_translate_property(property):
        return properties_to_translate.get(property)

    current_value = keyword = None

    for lineno, line in enumerate(fileobj, start=1):
        line = line.decode(encoding)

        if current_value:
            remainder = current_value.parse_line(line)
            if remainder is None:
                pass  # Still un-terminated
            elif remainder.strip():
                raise ValueError("Trailing data after string")
            else:
                for value, value_lineno in current_value.get_result():
                    yield (
                        value_lineno,
                        keyword,
                        [value],
                        [],
                    )
                current_value = None
            continue

        if line.startswith('['):
            continue

        match = _godot_property_str.match(line)
        if match:
            property = match.group(1)
            value = match.group(2)
            keyword = check_translate_property(property)
            if keyword:
                if value[0:1] == '[':
                    current_value = ArrayReader(lineno)
                else:
                    current_value = StringReader(lineno)
                remainder = current_value.parse_line(value[1:])
                if remainder is None:
                    pass  # Un-terminated string
                elif not remainder.strip():
                    for value, value_lineno in current_value.get_result():
                        yield (value_lineno, keyword, [value], [])
                    current_value = None
                else:
                    raise ValueError("Trailing data after string")
